(*
   The lexical analyzer. The file lexer.ml is generated automatically
   from lexer.mll.  The only modification commonly needed here is
   adding new keywords to the list of reserved words at the top.
*)

{
open Lolevel.Error

let reserved_words = [
  (* Keywords *)
  ("symbol", fun i -> Parser.SYMBOL i);
  ("if", fun i -> Parser.IF i);
  ("then", fun i -> Parser.THEN i);
  ("else", fun i -> Parser.ELSE i);
  ("lambda", fun i -> Parser.LAMBDA i);
  ("let", fun i -> Parser.LET i);
  ("letrec", fun i -> Parser.LETREC i);
  ("in", fun i -> Parser.IN i);
  ("end", fun i -> Parser.END i);

  (* We declare these without the parser using them so that they can't
     be used for variables. *)
  ("match", fun i -> Parser.MATCH i);
  ("case", fun i -> Parser.CASE i);
  ("of", fun i -> Parser.OF i);
  ("when", fun i -> Parser.WHEN i);
  ("as", fun i -> Parser.AS i);

  ("true", fun i -> Parser.TRUE i);
  ("false", fun i -> Parser.FALSE i);
  ("succ", fun i -> Parser.SUCC i);
  ("box", fun i -> Parser.BOX i);
  ("pred", fun i -> Parser.PRED i);
  ("zero?", fun i -> Parser.ISZERO i);
  ("not", fun i -> Parser.NOT i);
  ("fst", fun i -> Parser.FST i);
  ("snd", fun i -> Parser.SND i);
  ("trd", fun i -> Parser.TRD i);

  ("and", fun i -> Parser.AND i);
  ("or", fun i -> Parser.OR i);
  ("id", fun i -> Parser.IDENTITY i);
  ("comp", fun i -> Parser.COMP i);
  ("pmoc", fun i -> Parser.PMOC i);
  ("const", fun i -> Parser.CONST i);
  ("flip", fun i -> Parser.FLIP i);
(*
  ("pair", fun i -> Parser.PAIR i);
  ("triple", fun i -> Parser.TRIPLE i);
  ("head", fun i -> Parser.HEAD i);
  ("tail", fun i -> Parser.TAIL i);
  ("cons", fun i -> Parser.CONS i);
  ("nil", fun i -> Parser.NIL i);
  ("swap", fun i -> Parser.SWAP i);
*)

  (* Symbols *)
  ("~", fun i -> Parser.TILDE i);
  ("`", fun i -> Parser.BACKQUOTE i); (* or GRAVE *)
  ("!", fun i -> Parser.BANG i);
  ("@", fun i -> Parser.AT i);
  ("#", fun i -> Parser.HASH i); (* or SHARP *)
  ("$", fun i -> Parser.DOLLAR i);
  ("%", fun i -> Parser.PERCENT i);
  ("^", fun i -> Parser.CARET i);
  ("&", fun i -> Parser.AMPER i);
  ("*", fun i -> Parser.STAR i);
  ("(", fun i -> Parser.LPAREN i);
  (")", fun i -> Parser.RPAREN i);
  ("_", fun i -> Parser.UNDER i);
  ("-", fun i -> Parser.MINUS i);
  ("+", fun i -> Parser.PLUS i);
  ("=", fun i -> Parser.EQUAL i);
  ("{", fun i -> Parser.LCURLY i);
  ("[", fun i -> Parser.LSQUARE i);
  ("}", fun i -> Parser.RCURLY i);
  ("]", fun i -> Parser.RSQUARE i);
  ("|", fun i -> Parser.BAR i);
  ("\\", fun i -> Parser.BACKSLASH i);
  (":", fun i -> Parser.COLON i);
  (";", fun i -> Parser.SEMI i);
  ("\"", fun i -> Parser.DQUOTE i);
  ("'", fun i -> Parser.APOSTROPHE i); (* or QUOTE *)
  ("<", fun i -> Parser.LESS i);
  (",", fun i -> Parser.COMMA i);
  (">", fun i -> Parser.GREATER i);
  (".", fun i -> Parser.DOT i);
  ("?", fun i -> Parser.QUESTION i);
  ("/", fun i -> Parser.SLASH i);

  (* Special compound symbols.
     Currently all of ( ) [ ] { } , ; ' ` and \ are treated
     as atomic by the lexer.  Other punctuation sequences will
     be treated as a Parser.LCID.
  *)
  (">=", fun i -> Parser.GREATEREQ i);
  ("<=", fun i -> Parser.LESSEQ i);
  ("==", fun i -> Parser.EQUALEQ i);


(*
  (":=", fun i -> Parser.COLONEQ i);
  ("!=", fun i -> Parser.BANGEQ i);
  ("+=", fun i -> Parser.PLUSEQ i);
  ("-=", fun i -> Parser.MINUSEQ i);
  ("*=", fun i -> Parser.STAREQ i);
  ("/=", fun i -> Parser.SLASHEQ i);
  ("~=", fun i -> Parser.TILDEEQ i);
  ("++", fun i -> Parser.PLUSPLUS i);
  ("--", fun i -> Parser.MINUSMINUS i);
  ("&&", fun i -> Parser.AMPERAMPER i);
  ("||", fun i -> Parser.BARBAR i);
  (";;", fun i -> Parser.SEMISEMI i);
  ("..", fun i -> Parser.DOTDOT i);
  ("::", fun i -> Parser.COLONCOLON i);
  ("<-", fun i -> Parser.BACKARROW i); (* also LESSMINUS *)
  ("->", fun i -> Parser.ARROW i); (* also MINUSGREATER *)
  ("=>", fun i -> Parser.DARROW i);
  ("==>", fun i -> Parser.DDARROW i);

  ("(|", fun i -> Parser.LPARENBAR i);
  ("|)", fun i -> Parser.BARRPAREN i);
  ("[|", fun i -> Parser.LSQUAREBAR i);
  ("|]", fun i -> Parser.BARRSQUARE i);
  ("{|", fun i -> Parser.LCURLYBAR i);
  ("|}", fun i -> Parser.BARRCURLY i);
  ("<|", fun i -> Parser.LESSBAR i);
  ("|>", fun i -> Parser.BARGREATER i);
  ("[<", fun i -> Parser.LSQUARELESS i);
  (">]", fun i -> Parser.GREATERRSQUARE i);
  ("{<", fun i -> Parser.LCURLYLESS i);
  (">}", fun i -> Parser.GREATERRCURLY i);
  ("<:", fun i -> Parser.LESSCOLON i);
  (":>", fun i -> Parser.COLONGREATER i);
  ("[:", fun i -> Parser.LSQUARECOLON i);
  (":]", fun i -> Parser.COLONRSQUARE i);
  ("{:", fun i -> Parser.LCURLYCOLON i);
  (":}", fun i -> Parser.COLONRCURLY i);

  ("><", fun i -> Parser.GTLT i);
  ("<>", fun i -> Parser.LTGT i);
  (">>", fun i -> Parser.GG i);
  (">>=", fun i -> Parser.GGE i);
  (">=>", fun i -> Parser.GEG i);
  ("<<", fun i -> Parser.LL i);
  ("=<<", fun i -> Parser.ELL i);
  ("<=<", fun i -> Parser.LEL i);
*)
]

(* Support functions *)
type symbol_builder = info -> Parser.token
let (symbol_table : (string, symbol_builder) Hashtbl.t) = Hashtbl.create 1024
let _ = List.iter (fun (str,f) -> Hashtbl.add symbol_table str f) reserved_words

let create_ident i str =
  try (Hashtbl.find symbol_table str) i
  with Not_found ->
    if (String.get str 0) >= 'A' && (String.get str 0) <= 'Z'
    then Parser.UCID {i=i;v=str}
    else Parser.LCID {i=i;v=str} (* includes unregistered punctuation/operator symbols *)

let from_channel infile chan =
  let filename = if not (Filename.is_implicit infile) then infile
                 else Filename.concat (Sys.getcwd()) infile in
  let lexbuf = Lexing.from_channel chan in
  let pos = lexbuf.Lexing.lex_curr_p in
  let pos' = { pos with Lexing.pos_fname = filename } in
  let () = lexbuf.Lexing.lex_curr_p <- pos' in
  let () = lexbuf.Lexing.lex_start_p <- pos' in
  lexbuf

let from_string str =
  let lexbuf = Lexing.from_string str in
  lexbuf

let text = Lexing.lexeme (* lexbuf *)
let char = Lexing.lexeme_char (* lexbuf idx *)
let info = Lexing.lexeme_start_p (* lexbuf *)
let newline = Lexing.new_line (* lexbuf *)

let string_buffer = ref (String.create 2048)
let string_end = ref 0

let reset_str () = string_end := 0

let add_str ch =
  let x = !string_end in
  let buffer = !string_buffer in
  if x = String.length buffer
  then begin
    let new_buffer = String.create (x*2) in
    String.blit buffer 0 new_buffer 0 x;
    String.set new_buffer x ch;
    string_buffer := new_buffer;
    string_end := x+1
  end
  else begin
    String.set buffer x ch;
    string_end := x+1
  end

let get_str () = String.sub (!string_buffer) 0 (!string_end)

}


(* Regex abbreviations *)

let digit = ['0'-'9']
let hexdigit = ['0'-'9' 'A'-'F' 'a'-'f']

(* horiz space: space \t \f and &nbsp; *)
(* omitted: \b \v \127 \0 \e=\027=\x1b *)
let blank = [' ' '\009' '\012' '\160']


(* Main body of the lexical analyzer *)

rule main = parse

| blank+            { main lexbuf }

| blank* '\r'? '\n' { newline lexbuf; main lexbuf }

| "*/" { error_with (info lexbuf) "Unmatched end of comment" }

| "/*" { comment 1 (info lexbuf) lexbuf; main lexbuf }

| digit+ '.' ['a'-'z' 'A'-'Z' '_' '.']
| digit+ ['a'-'z' 'A'-'Z' '_'] { error_with (info lexbuf) "Invalid number" }

| digit+ '.' { let initial = String.sub (text lexbuf) 0 (String.length (text lexbuf) - 1) in
               Parser.DOTV{i=info lexbuf; v=int_of_string(initial)} }

| digit+ { Parser.INTV{i=info lexbuf; v=int_of_string (text lexbuf)} }

| ['A'-'Z' 'a'-'z'] ['A'-'Z' 'a'-'z' '_' '0'-'9']* ('?' | '!' | '\''*)
    { create_ident (info lexbuf) (text lexbuf) }

(* unrepeatable punctuation *)
| ['(' ')' '[' ']' '{' '}' ',' ';' '\'' '`' '\\']
    { create_ident (info lexbuf) (text lexbuf) }

(* repeatable punctuation, omitting '!' '?' *)
| ['~' '@' '#' '$' '%' '^' '&' '*' '-' '+' '=' '_' '|' ':' '<' '>' '.' '/']+
    { create_ident (info lexbuf) (text lexbuf) }

| '"' { reset_str(); string (info lexbuf) lexbuf }

| eof { Parser.EOF (info lexbuf) }

| _ { error_with (info lexbuf) "Illegal character" }

and comment depth start_info = parse
| "/*"
    { comment (depth + 1) start_info lexbuf }
| "*/"
    { if depth > 1 then comment (depth - 1) start_info lexbuf }
| eof
    { error_with start_info "Comment not terminated" }
| [^ '\n']
    { comment depth start_info lexbuf }
| '\n'
    { newline lexbuf; comment depth start_info lexbuf }

and string start_info = parse
| '"'  { Parser.STRINGV {i = start_info; v=get_str()} }
| '\\' { add_str (escaped lexbuf); string start_info lexbuf }
| '\n' { add_str '\n'; newline lexbuf; string start_info lexbuf }
| eof  { error_with start_info "String not terminated" }
| _    { add_str (char lexbuf 0); string start_info lexbuf }

and escaped = parse
| 'n'       { '\n' }
| 't'       { '\t' }
| '\\'      { '\\' }
| '"'       { '\034' }
| '\''      { '\'' }
| ['x' 'X'] hexdigit hexdigit
    {
      let hi = char lexbuf 1 in
      let lo = char lexbuf 2 in
      let hi' = Char.code hi - (if hi >= 'a' then 87 else if hi >= 'A' then 55 else 48) in
      let lo' = Char.code lo - (if lo >= 'a' then 87 else if lo >= 'A' then 55 else 48) in
      Char.chr (hi'*16 + lo')
    }
| digit digit digit
    {
      let byte = int_of_string (text lexbuf) in
      if byte > 255 then error_with (info lexbuf) "Illegal character constant"
      else Char.chr byte
    }
| _    (* [^ '"' '\\' 't' 'n' '\''] *)
    { error_with (info lexbuf) "Illegal character constant" }

(*  *)
